/*
** ADM-XRC Series Driver
**
** (C) Copyright 2001-2004 Alpha Data
**
** intwait.c - IOCTL handlers for waiting on FPGA interrupts.
**
** ----------------------------------------------------------------------------
*/

#include "platform.h"
#include "device.h"
#include "ioctl_code.h"
#include "ioctls.h"
#include "intwait.h"

typedef enum _State {
  StateIdle = 0,
  StateWaiting = 1,
  StateCompleting = 2
} State;

static void
intWaitComplete(
  DeviceContext* pDevCtx,
  unsigned int index)
{
  OS_SPINLOCK_FLAGS f;
  os_boolean bComplete;
  IntWaitNode* node;
  IntWaitDeviceContext* pContext;

  os_dbgprint(3, ("intWaitComplete: entered, index=%u\n", index));

  pContext = &pDevCtx->intwait[index];
  os_acquire_spinlock(&pContext->lock, &f);
  node = pContext->waiting;
  if (node == NULL) {
    /* No IO request waiting - set the signalled flag and do nothing more */
    os_dbgprint(6, ("intWaitComplete: no IO request queued, setting signalled flag\n"));
    pContext->signalled = TRUE;
    bComplete = FALSE;
  } else {
    pContext->waiting = NULL;
    node->state = StateCompleting;
    bComplete = TRUE;
  }
  os_release_spinlock(&pContext->lock, f);

  if (bComplete) {
    os_signal_event(&node->event);
  }
}

static void
signalIntEvents(
  DeviceContext* pDevCtx,
  unsigned int index)
{
  OS_SPINLOCK_FLAGS f;
  InterruptEvent* ev;
  IntWaitDeviceContext* pContext;

  /* Signal all registered interrupt wait events */
  pContext = &pDevCtx->intwait[index];
  os_acquire_spinlock(&pContext->lock, &f);
  ev = pContext->eventList;
  while (1) {
    if (ev == NULL) {
      break;
    }
    os_dbgprint(6, ("signalIntEvents: event list node = 0x%p\n", ev));
    os_set_event(ev->event);
    ev = ev->next;
  }
  os_release_spinlock(&pContext->lock, f);
}

void
ADMXRC_FpgaInterrupt(
  DeviceContext* pDevCtx,
  unsigned int index)
{
  intWaitComplete(pDevCtx, index);
  signalIntEvents(pDevCtx, index);
}

/*
** IOCTL handlers
*/

OS_STATUS
ADMXRC_CancelWaitForInterrupt(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* pReq)
{
  OS_SPINLOCK_FLAGS f;
  os_boolean bComplete;
  IntWaitNode* node = NULL;
  OS_STATUS status = OS_STATUS_SUCCESS;
  IntWaitDeviceContext* pContext;

  os_dbgprint(3, ("ADMXRC_CancelWaitForInterrupt: entered\n"));

  pContext = &pDevCtx->intwait[0];
  os_acquire_spinlock(&pContext->lock, &f);
  node = pContext->waiting;
  if (node == NULL) {
    bComplete = FALSE;
  } else {
    node->state = StateCompleting;
    node->bCancel = TRUE;
    bComplete = TRUE;
  }
  os_release_spinlock(&pContext->lock, f);

  if (bComplete) {
    os_signal_event(&node->event);
  } else {
    status = OS_MAKE_STATUS(ADMXRC2_FAILED);
  }

  return status;
}

OS_STATUS
ADMXRC_WaitForInterrupt(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* pReq,
  ADMXRC_IOCTLS_WAITFORINTERRUPT* pIoctl,
  unsigned long inSize,
  unsigned long outSize)
{
  OS_SPINLOCK_FLAGS f;
  OS_STATUS status = OS_STATUS_SUCCESS;
  IntWaitNode* node = NULL;
  os_ticks start, now, expire, ticks;
  unsigned long remaining_us;
  os_boolean bSignalled, bCancelled;
  os_boolean bUseTimeout;
  IntWaitDeviceContext* pContext;

  if (pIoctl == NULL || inSize != sizeof(pIoctl->in) || outSize != sizeof(pIoctl->out)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  pContext = &pDevCtx->intwait[0];

  /* Get values indicating whether or not to use a timeout */
  bUseTimeout = (pIoctl->in.flags & ADMXRC_INTWAIT_TIMEOUT) != 0;
  if (bUseTimeout) {
    start = os_get_time();
    expire = start + os_us_to_ticks(pIoctl->in.timeout);
  } else {
    expire = 0; /* Stop compiler complaining about uninit vars. */
  }

  node = (IntWaitNode*) os_alloc_lcache(pContext->ctlPool, 0);
  os_init_event(&node->event);
  node->bCancel = FALSE;
  node->state = StateWaiting;
  node->pReq = pReq;

  os_acquire_spinlock(&pContext->lock, &f);
  if (pContext->waiting != NULL) {
    /* A thread is already waiting for an interrupt */
    os_release_spinlock(&pContext->lock, f);
    status = OS_MAKE_STATUS(ADMXRC2_DEVICE_BUSY);
  } else {
    if (pContext->signalled) {
      /* An FPGA interrupt has already occurred */
      pContext->signalled = FALSE;
      os_release_spinlock(&pContext->lock, f);
      pIoctl->out.remaining = pIoctl->in.timeout;
      pReq->status.length = sizeof(pIoctl->out);
      status = OS_STATUS_SUCCESS;
    } else {
      pContext->waiting = node;
      os_release_spinlock(&pContext->lock, f);
      if (bUseTimeout) {
        now = os_get_time();
        if (os_time_after_eq(now, expire)) {
          ticks = 0;
        } else {
          ticks = expire - now;
        }
        bCancelled = os_wait_event_timeout(&node->event, ticks, TRUE, &bSignalled);
        if (bCancelled) {
          if (bSignalled) {
            status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
          } else {
            status = OS_MAKE_STATUS(ADMXRC2_TIMEOUT);
          }
        } else {
          if (node->bCancel) {
            status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
          } else {
            status = OS_STATUS_SUCCESS;
          }
        }
      } else {
        bSignalled = os_wait_event(&node->event, TRUE);
        if (bSignalled) {
          status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
        } else {
          if (node->bCancel) {
            status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
          } else {
            status = OS_STATUS_SUCCESS;
          }
        }
      }
      os_acquire_spinlock(&pContext->lock, &f);
      pContext->waiting = NULL;
      os_release_spinlock(&pContext->lock, f);
    }
  }

  if (OS_SUCCESS(status)) {
    pReq->status.length = sizeof(pIoctl->out);
    if (bUseTimeout) {
      now = os_get_time();
      if (os_time_after_eq(now, expire)) {
        pIoctl->out.remaining = 0;
      } else {
        remaining_us = os_ticks_to_us(expire - now);
        if (remaining_us > pIoctl->in.timeout) {
          pIoctl->out.remaining = pIoctl->in.timeout;
        } else {
          pIoctl->out.remaining = remaining_us;
        }
      }
    }
  }

  os_uninit_event(&node->event);
  if (node != NULL) {
    os_free_lcache(pContext->ctlPool, node);
  }

  return status;
}

void
ADMXRC_UnregAllInterruptEvents(
  DeviceContext* pDevCtx)
{
  OS_SPINLOCK_FLAGS f;
  InterruptEvent* node;
  InterruptEvent* next;
  int n;
  IntWaitDeviceContext* pContext;

  os_dbgprint(2, ("AdmXrc_UnregAllIntEvents: entered\n"));

  pContext = &pDevCtx->intwait[0];

  os_acquire_spinlock(&pContext->lock, &f);
  node = pContext->eventList;
  pContext->eventList = NULL;
  os_release_spinlock(&pContext->lock, f);

  n = 0;
  while (node != NULL) {
    next = node->next;
    os_kernel_free(node);
    node = next;
    n++;
  }

  os_dbgprint(2, ("AdmXrc_UnregAllIntEvents: %ld interrupt events unregistered\n", 
    (unsigned long) n));
}

OS_STATUS
ADMXRC_RegInterruptEvent(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* pReq,
  ADMXRC_IOCTLS_REGINTERRUPTEVENT* pIoctl,
  unsigned long inSize)
{

  OS_SPINLOCK_FLAGS f;
  InterruptEvent* node;
  OS_STATUS status = OS_STATUS_SUCCESS;
  IntWaitDeviceContext* pContext;

  if (pIoctl == NULL || inSize != sizeof(pIoctl->in)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  pContext = &pDevCtx->intwait[0];

  node = (InterruptEvent*) os_kernel_malloc(sizeof(InterruptEvent));
  if (node == NULL) {
    return OS_MAKE_STATUS(ADMXRC2_NO_MEMORY);
  } 
  node->event = pIoctl->in.event;

  os_acquire_spinlock(&pContext->lock, &f);
  node->next = pContext->eventList;
  pContext->eventList = node;
  os_release_spinlock(&pContext->lock, f);

  return status;
}

OS_STATUS
ADMXRC_UnregInterruptEvent(
                           DeviceContext* pDevCtx,
                           OS_IO_REQUEST* pReq,
                           ADMXRC_IOCTLS_UNREGINTERRUPTEVENT* pIoctl,
                           unsigned long inSize)
{
  OS_SPINLOCK_FLAGS f;
  InterruptEvent* node;
  InterruptEvent** pnode;
  os_event event;
  IntWaitDeviceContext* pContext;

  if (pIoctl == NULL || inSize != sizeof(pIoctl->in)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  pContext = &pDevCtx->intwait[0];

  event = pIoctl->in.event;
  os_acquire_spinlock(&pContext->lock, &f);
  pnode = &pContext->eventList;
  while (1) {
    node = *pnode;
    if (node == NULL) {
      break;
    } else {
      if (node->event == event) {
        *pnode = node->next;
        break;
      } else {
        pnode = &node->next;
      }
    }
  }
  os_release_spinlock(&pContext->lock, f);

  if (node == NULL) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_HANDLE);
  } else {
    os_kernel_free(node);
  }

  return OS_STATUS_SUCCESS;
}
