/*
** ADM-XRC Series Driver
**
** (C) Copyright 2001-2004 Alpha Data
**
** loadfpga.c - IOCTL handlers for configuring/unconfiguring the FPGA.
**
** ----------------------------------------------------------------------------
*/

#include "platform.h"
#include "device.h"
#include "drvparam.h"
#include "dmacore.h"
#include "dmadesc.h"
#include <ioctl_code.h>
#include "ioctls.h"
#include "loadfpga.h"

static OS_STATUS
clearFPGA(
  DeviceContext* pDevCtx,
  unsigned int index)
{
  OS_STATUS status;
  unsigned long timeout;
  unsigned long now;
  os_boolean bSuccess;
  os_boolean bInterrupt;
  os_ticks delay;
  os_boolean bVal;

  /*
  ** Pulse PROG for 20 usec - this will cause DONE to
  ** become deasserted and INIT to (eventually) become
  ** deasserted.
  */
  bVal = TRUE;
  pDevCtx->methods.fpga(pDevCtx, index, MethodFpgaOp_SetProg, &bVal);
  os_delay_busy_us(20);
  bVal = FALSE;
  pDevCtx->methods.fpga(pDevCtx, index, MethodFpgaOp_SetProg, &bVal);
  
  /*
  ** Delay until DONE is deasserted.
  **
  ** We allow a timeout of 1s.
  */
  timeout = os_get_time() + os_ticks_per_second;
  bSuccess = FALSE;
  while (1) {
    pDevCtx->methods.fpga(pDevCtx, index, MethodFpgaOp_GetDone, &bVal);
    if (!bVal) {
      bSuccess = TRUE;
      break;
    }
    
    now = os_get_time();
    if (now >= timeout) {
      os_dbgprint(4, ("clearFPGA: timeout waiting for DONE deasserted\n"));
      /* Timeout occurred */
      break;
    }

    delay = 10;
    bInterrupt = os_delay_thread(TRUE, &delay);
    if (bInterrupt) {
      os_dbgprint(3, ("clearFPGA: interrupt while waiting for DONE deasserted\n"));
      status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
      goto done;
    }
  }
  
  if (!bSuccess) {
    status = OS_MAKE_STATUS(ADMXRC2_FAILED);
    goto done;
  }

  /*
  ** Delay until INIT is deasserted.
  **
  ** We allow a timeout of 1s.
  */
  timeout = os_get_time() + os_ticks_per_second;
  bSuccess = FALSE;
  while (1) {
    pDevCtx->methods.fpga(pDevCtx, index, MethodFpgaOp_GetInit, &bVal);
    if (!bVal) {
      bSuccess = TRUE;
      break;
    }
    
    now = os_get_time();
    if (now >= timeout) {
      os_dbgprint(4, ("clearFPGA: timeout waiting for INIT deasserted\n"));
      /* Timeout occurred */
      break;
    }

    delay = 10;
    bInterrupt = os_delay_thread(TRUE, &delay);
    if (bInterrupt) {
      os_dbgprint(3, ("clearFPGA: interrupt while waiting for INIT deasserted\n"));
      status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
      goto done;
    }
  }
  
  if (!bSuccess) {
    status = OS_MAKE_STATUS(ADMXRC2_FAILED);
    goto done;
  }
  
  status = OS_STATUS_SUCCESS;
  
 done:
  return status;
}

static OS_STATUS
checkConfigured(
  DeviceContext* pDevCtx,
  unsigned int index)
{
  OS_STATUS status = OS_STATUS_SUCCESS;
  os_boolean bSuccess;
  unsigned long timeout;
  unsigned long now;
  os_boolean bInterrupt;
  os_ticks delay;
  os_boolean bVal;

  /*
  ** Wait for device to assert DONE, which may rely on DLL's/DCM's.
  **
  ** We allow a timeout of 1s.
  */
  timeout = os_get_time() + os_ticks_per_second;
  bSuccess = FALSE;
  while (1)	{
    pDevCtx->methods.fpga(pDevCtx, index, MethodFpgaOp_GetDone, &bVal);
    if (bVal) {
      bSuccess = TRUE;
      break;
    }
    
    now = os_get_time();
    if (now >= timeout) {
      os_dbgprint(4, ("clearFPGA: timeout waiting for DONE asserted\n"));
      /* Timeout occurred */
      break;
    }

    delay = 10;
    bInterrupt = os_delay_thread(TRUE, &delay);
    if (bInterrupt) {
      os_dbgprint(3, ("checkConfigured: interrupt while waiting for DONE asserted\n"));
      status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
      goto done;
    }
  }
  
  if (!bSuccess) {
    pDevCtx->methods.fpga(pDevCtx, index, MethodFpgaOp_GetInit, &bVal);
    if (bVal) {
      os_dbgprint(3, ("checkConfigured: Init indicates failure to load\n"));
    } else {
      os_dbgprint(3, ("checkConfigured: Done did not go high\n"));
    }
    
    /*
    ** Something bad happened, so clear the FPGA in case
    ** it has been loaded with a bad bitstream.
    */
    bVal = TRUE;
    pDevCtx->methods.fpga(pDevCtx, index, MethodFpgaOp_SetProg, &bVal);
    os_delay_busy_us(20);
    bVal = FALSE;
    pDevCtx->methods.fpga(pDevCtx, index, MethodFpgaOp_SetProg, &bVal);
    status = OS_MAKE_STATUS(ADMXRC2_FAILED);
  }
  
 done:
  return status;
}

/*
** IOCTL handlers
*/

OS_STATUS
ADMXRC_LoadFPGA(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* pReq)
{
  os_uint8* data;
  unsigned long length;
  OS_STATUS status;
  OS_SPINLOCK_FLAGS f;

  os_dbgprint(3, ("ADMXRC_LoadFPGA: entered\n"));

  data = (os_uint8*) pReq->param.massBuf;
  length = pReq->param.massSize;

  os_dbgprint(3, ("ADMXRC_LoadFPGA: data = 0x%p length = %ld\n", (void*) data, length));

  os_acquire_spinlock(&pDevCtx->fpga.lock, &f);
  if (pDevCtx->fpga.busy[0]) {
    os_release_spinlock(&pDevCtx->fpga.lock, f);
    return OS_MAKE_STATUS(ADMXRC2_DEVICE_BUSY);
  } else {
    pDevCtx->fpga.busy[0] = TRUE;
    os_release_spinlock(&pDevCtx->fpga.lock, f);
  }
  
  status = clearFPGA(pDevCtx, 0);
  if (!OS_SUCCESS(status)) {
    os_dbgprint(3, ("ADMXRC_LoadFPGA: failed to clear FPGA\n"));
    goto done;
  }
  
  os_dbgprint(3, ("ADMXRC_LoadFPGA: cleared FPGA\n"));

  /* Send the bitstream to the FPGA */
  pDevCtx->methods.selectMap(pDevCtx, 0, TRUE, data++, length);
  os_dbgprint(3, ("ADMXRC_LoadFPGA: sent bitstream to FPGA\n"));
  
  status = checkConfigured(pDevCtx, 0);
  if (!OS_SUCCESS(status)) {
    os_dbgprint(3, ("ADMXRC_LoadFPGA: configuration failed\n"));
    goto done;
  }
  
  os_dbgprint(3, ("ADMXRC_LoadFPGA: success\n"));

  status = OS_STATUS_SUCCESS;
  
 done:
  os_acquire_spinlock(&pDevCtx->fpga.lock, &f);
  pDevCtx->fpga.busy[0] = FALSE;
  os_release_spinlock(&pDevCtx->fpga.lock, f);
  
  return status;
}

OS_STATUS
ADMXRC_LoadFPGAExt(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* pReq,
  ADMXRC_IOCTLS_LOADFPGAEXT* pIoctl,
  unsigned long inSize)
{
  os_uint8* data;
  unsigned long length;
  OS_STATUS status;
  OS_SPINLOCK_FLAGS f;
  unsigned int index;

  os_dbgprint(3, ("ADMXRC_LoadFPGAExt: entered\n"));

  if (pIoctl == NULL || inSize != sizeof(*pIoctl)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  data = (os_uint8*) pReq->param.massBuf;
  length = pReq->param.massSize;
  index = pIoctl->in.index;

  os_dbgprint(3, ("ADMXRC_LoadFPGAExt: index=%ld data=0x%p length=%ld\n",
		  (unsigned long)index, (void*) data, length));

  if (index >= pDevCtx->miscInfo.numFpga) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  os_acquire_spinlock(&pDevCtx->fpga.lock, &f);
  if (pDevCtx->fpga.busy[index]) {
    os_release_spinlock(&pDevCtx->fpga.lock, f);
    return OS_MAKE_STATUS(ADMXRC2_DEVICE_BUSY);
  } else {
    pDevCtx->fpga.busy[index] = TRUE;
    os_release_spinlock(&pDevCtx->fpga.lock, f);
  }
  
  status = clearFPGA(pDevCtx, index);
  if (!OS_SUCCESS(status)) {
    os_dbgprint(3, ("ADMXRC_LoadFPGAExt: failed to clear FPGA\n"));
    goto done;
  }
  
  os_dbgprint(3, ("ADMXRC_LoadFPGA: cleared FPGA\n"));

  /* Send the bitstream to the FPGA */
  pDevCtx->methods.selectMap(pDevCtx, index, TRUE, data++, length);
  
  status = checkConfigured(pDevCtx, index);
  if (!OS_SUCCESS(status)) {
    os_dbgprint(3, ("ADMXRC_LoadFPGAExt: configuration failed\n"));
    goto done;
  }
  
  os_dbgprint(3, ("ADMXRC_LoadFPGAExt: success\n"));

  status = OS_STATUS_SUCCESS;
  
 done:
  os_acquire_spinlock(&pDevCtx->fpga.lock, &f);
  pDevCtx->fpga.busy[index] = FALSE;
  os_release_spinlock(&pDevCtx->fpga.lock, f);
  
  return status;
}

static void
loadFPGADMACallback(
  DeviceContext* pDevCtx,
  DmaNode* node)
{
  OS_EVENT* pEvent;
  
  os_dbgprint(6, ("loadFPGADMACallback\n"));
  
  pEvent = (OS_EVENT*) node->client.pContext;
  os_signal_event(pEvent);
}

OS_STATUS
ADMXRC_LoadFPGADMA(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* pReq,
  ADMXRC_IOCTLS_LOADFPGADMA* pIoctl,
  unsigned long inSize)
{
  void* buffer;
  unsigned long length;
  OS_STATUS status;
  OS_SPINLOCK_FLAGS f;
  DmaNodeLoadFpga* pControl = NULL;
  DmaNode* pDmaNode = NULL;
  DmaTarget* pDmaTgt = NULL;
  os_uint32 mode;
  os_uint32 local;
  unsigned long flags;
  unsigned char channel, queue;
  os_boolean locked = FALSE;
  OS_EVENT event;
  os_boolean bSignalled, bMustStart;

  os_dbgprint(3, ("ADMXRC_LoadFPGADMA: entered\n"));

  if (pIoctl == NULL || inSize != sizeof(*pIoctl)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  flags = pIoctl->in.flags;
  if (flags & ~(ADMXRC2_DMAFLAG_TIMEOUT | ADMXRC2_DMAFLAG_DONOTQUEUE)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }
  flags &= ~ADMXRC_DMAFLAG_TIMEOUT;

  channel = pIoctl->in.channel;
  if (channel != ADMXRC2_DMACHAN_ANY && channel >= pDevCtx->info.cardInfo.numDmaChan) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  length = pIoctl->in.length;
  buffer = pIoctl->in.buffer;
  if (buffer == NULL) {
    return OS_MAKE_STATUS(ADMXRC2_NULL_POINTER);
  }

  switch (pDevCtx->info.cardInfo.boardType) {
  case ADMXRC2_BOARD_ADMXRC4LS:
  case ADMXRC2_BOARD_ADMXRC4LX:
  case ADMXRC2_BOARD_ADMXRC4SX:
    return OS_MAKE_STATUS(ADMXRC2_NOT_SUPPORTED);
  default:
    break;
  }

  os_init_event(&event);

  /* Lock down the user buffer and obtain an MDL for it */
  status = ADMXRC_CreateDMATarget(pDevCtx, buffer, length, &pDmaTgt);
  if (!OS_SUCCESS(status)) {
    goto completed;
  }

  pControl = os_alloc_lcache(pDevCtx->fpga.ctlPool, 0);
  if (pControl == NULL) {
    status = OS_MAKE_STATUS(ADMXRC2_NO_MEMORY);
    goto completed;
  }

  os_acquire_spinlock(&pDevCtx->fpga.lock, &f);
  if (pDevCtx->fpga.busy[0]) {
    os_release_spinlock(&pDevCtx->fpga.lock, f);
    status = OS_MAKE_STATUS(ADMXRC2_DEVICE_BUSY);
    goto completed;
  } else {
    locked = TRUE;
    pDevCtx->fpga.busy[0] = TRUE;
    os_release_spinlock(&pDevCtx->fpga.lock, f);
  }

  status = clearFPGA(pDevCtx, 0);
  if (!OS_SUCCESS(status)) {
    goto completed;
  }

  /* Determine which DMA queue to add this node to */
  queue = (channel == ADMXRC2_DMACHAN_ANY) ? pDevCtx->info.cardInfo.numDmaChan : (unsigned char) channel;
  
  /* Calculate local bus address of SelectMap port and DMA mode */
  local = pDevCtx->info.spaceInfo[1].localBase + pDevCtx->miscInfo.selectMapOffset[0];
  local += (pDevCtx->hw.control.byte - (os_uint8*) pDevCtx->hw.space1);
  if (pDevCtx->miscInfo.bSelectMap32Bit) {
    mode = 0x000009c2U;
  } else {
    mode = 0x000009c0U;
  }
  
  pDmaNode = &pControl->dmanode;
  pDmaNode->client.pContext = (void*)&event;
  pDmaNode->client.pCallback = loadFPGADMACallback;
  pDmaNode->client.pBuffer = pDmaTgt;
  pDmaNode->client.direction = (unsigned char)ADMXRC2_PCITOLOCAL;
  pDmaNode->client.length = length;
  pDmaNode->client.offset = 0;
  pDmaNode->client.mode = mode;
  pDmaNode->client.local = local;
  pDmaNode->client.queue = queue;
  pDmaNode->client.bTwoD = FALSE;
  
  os_dbgprint(6, ("ADMXRC_LoadFPGADMA: enqueuing DMANODE 0x%p buffer=0x%p len=0x%ld loc=0x%lx dir=%ld chan=%ld mode=0x%lx flags=0x%lx q=%ld\n",
		  (void*)pDmaNode,
		  (void*)buffer,
		  (unsigned long)length,
		  (unsigned long)local,
		  (unsigned long)ADMXRC2_PCITOLOCAL,
		  (unsigned long)channel,
		  (unsigned long)mode,
		  (unsigned long)flags,
		  (unsigned long)queue));

  status = OS_STATUS_SUCCESS;
  pReq->user.context = (void*)pControl;

  bMustStart = (flags & ADMXRC2_DMAFLAG_DONOTQUEUE) != 0;
  if (ADMXRC_EnqueueDma(pDevCtx, pDmaNode, bMustStart)) {
    status = OS_MAKE_STATUS(ADMXRC2_DEVICE_BUSY);
    goto completed;
  }

  bSignalled = os_wait_event(&event, TRUE);
  if (bSignalled) {
    if (!ADMXRC_CancelDma(pDevCtx, pDmaNode)) {
      os_wait_event(&event, FALSE);
      status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
      goto completed;
    }
  }
  status = pControl->dmanode.status.status;

 completed:
  os_uninit_event(&event);

  if (OS_SUCCESS(status)) {
    status = checkConfigured(pDevCtx, 0);
  }

  if (locked) {
    os_acquire_spinlock(&pDevCtx->fpga.lock, &f);
    pDevCtx->fpga.busy[0] = FALSE;
    os_release_spinlock(&pDevCtx->fpga.lock, f);
  }
  
  if (pDmaTgt != NULL) {
    ADMXRC_ReleaseDMATarget(pDevCtx, pDmaTgt);
  }

  if (pControl != NULL) {
    os_free_lcache(pDevCtx->fpga.ctlPool, pControl);
  }

  return status;
}

OS_STATUS
ADMXRC_LoadFPGADMAExt(
  DeviceContext* pDevCtx,
  OS_IO_REQUEST* pReq,
  ADMXRC_IOCTLS_LOADFPGADMAEXT* pIoctl,
  unsigned long inSize)
{
  void* buffer;
  unsigned long length;
  OS_STATUS status;
  OS_SPINLOCK_FLAGS f;
  DmaNodeLoadFpga* pControl = NULL;
  DmaNode* pDmaNode = NULL;
  DmaTarget* pDmaTgt = NULL;
  os_uint32 mode;
  os_uint32 local;
  unsigned int index;
  unsigned long flags;
  unsigned char channel, queue;
  os_boolean locked = FALSE;
  OS_EVENT event;
  os_boolean bSignalled, bMustStart;

  os_dbgprint(3, ("ADMXRC_LoadFPGADMAExt: entered\n"));

  if (pIoctl == NULL || inSize != sizeof(*pIoctl)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  index = pIoctl->in.index;
  if (index >= pDevCtx->miscInfo.numFpga) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  flags = pIoctl->in.flags;
  if (flags & ~(ADMXRC2_DMAFLAG_TIMEOUT | ADMXRC2_DMAFLAG_DONOTQUEUE)) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }
  flags &= ~ADMXRC_DMAFLAG_TIMEOUT;

  channel = pIoctl->in.channel;
  if (channel != ADMXRC2_DMACHAN_ANY && channel >= pDevCtx->info.cardInfo.numDmaChan) {
    return OS_MAKE_STATUS(ADMXRC2_INVALID_PARAMETER);
  }

  length = pIoctl->in.length;
  buffer = pIoctl->in.buffer;
  if (buffer == NULL) {
    return OS_MAKE_STATUS(ADMXRC2_NULL_POINTER);
  }

  switch (pDevCtx->info.cardInfo.boardType) {
  case ADMXRC2_BOARD_ADMXRC4LS:
  case ADMXRC2_BOARD_ADMXRC4LX:
  case ADMXRC2_BOARD_ADMXRC4SX:
    return OS_MAKE_STATUS(ADMXRC2_NOT_SUPPORTED);
  default:
    break;
  }

  os_init_event(&event);

  /* Lock down the user buffer and obtain an MDL for it */
  status = ADMXRC_CreateDMATarget(pDevCtx, buffer, length, &pDmaTgt);
  if (!OS_SUCCESS(status)) {
    goto completed;
  }

  pControl = os_alloc_lcache(pDevCtx->fpga.ctlPool, 0);
  if (pControl == NULL) {
    status = OS_MAKE_STATUS(ADMXRC2_NO_MEMORY);
    goto completed;
  }

  os_acquire_spinlock(&pDevCtx->fpga.lock, &f);
  if (pDevCtx->fpga.busy[index]) {
    os_release_spinlock(&pDevCtx->fpga.lock, f);
    status = OS_MAKE_STATUS(ADMXRC2_DEVICE_BUSY);
    goto completed;
  } else {
    locked = TRUE;
    pDevCtx->fpga.busy[index] = TRUE;
    os_release_spinlock(&pDevCtx->fpga.lock, f);
  }

  status = clearFPGA(pDevCtx, index);
  if (!OS_SUCCESS(status)) {
    goto completed;
  }

  /* Determine which DMA queue to add this node to */
  queue = (channel == ADMXRC2_DMACHAN_ANY) ? pDevCtx->info.cardInfo.numDmaChan : (unsigned char) channel;
  
  /* Calculate local bus address of SelectMap port and DMA mode */
  local = pDevCtx->info.spaceInfo[1].localBase + pDevCtx->miscInfo.selectMapOffset[index];
  local += (pDevCtx->hw.control.byte - (os_uint8*) pDevCtx->hw.space1);
  if (pDevCtx->miscInfo.bSelectMap32Bit) {
    mode = 0x000009c2U;
  } else {
    mode = 0x000009c0U;
  }
  
  pDmaNode = &pControl->dmanode;
  pDmaNode->client.pContext = (void*) &event;
  pDmaNode->client.pCallback = loadFPGADMACallback;
  pDmaNode->client.pBuffer = pDmaTgt;
  pDmaNode->client.direction = (unsigned char) ADMXRC2_PCITOLOCAL;
  pDmaNode->client.length = length;
  pDmaNode->client.offset = 0;
  pDmaNode->client.mode = mode;
  pDmaNode->client.local = local;
  pDmaNode->client.queue = queue;
  pDmaNode->client.bTwoD = FALSE;
  
  os_dbgprint(6, ("ADMXRC_LoadFPGADMAExt: enqueuing DMANODE 0x%p buffer=0x%p len=0x%ld loc=0x%lx dir=%ld chan=%ld mode=0x%lx flags=0x%lx q=%ld\n",
		  (void*)pDmaNode,
		  (void*)buffer,
		  (unsigned long)length,
		  (unsigned long)local,
		  (unsigned long)ADMXRC2_PCITOLOCAL,
		  (unsigned long)channel,
		  (unsigned long)mode,
		  (unsigned long)flags,
		  (unsigned long)queue));

  status = OS_STATUS_SUCCESS;
  pReq->user.context = (void*) pControl;

  bMustStart = (flags & ADMXRC2_DMAFLAG_DONOTQUEUE) != 0;
  if (ADMXRC_EnqueueDma(pDevCtx, pDmaNode, bMustStart)) {
    status = OS_MAKE_STATUS(ADMXRC2_DEVICE_BUSY);
    goto completed;
  }

  bSignalled = os_wait_event(&event, TRUE);
  if (bSignalled) {
    if (!ADMXRC_CancelDma(pDevCtx, pDmaNode)) {
      os_wait_event(&event, FALSE);
      status = OS_MAKE_STATUS(ADMXRC2_CANCELLED);
      goto completed;
    }
  }
  status = pControl->dmanode.status.status;

 completed:
  os_uninit_event(&event);

  if (OS_SUCCESS(status)) {
    status = checkConfigured(pDevCtx, index);
  }

  if (locked) {
    os_acquire_spinlock(&pDevCtx->fpga.lock, &f);
    pDevCtx->fpga.busy[index] = FALSE;
    os_release_spinlock(&pDevCtx->fpga.lock, f);
  }
  
  if (pDmaTgt != NULL) {
    ADMXRC_ReleaseDMATarget(pDevCtx, pDmaTgt);
  }

  if (pControl != NULL) {
    os_free_lcache(pDevCtx->fpga.ctlPool, pControl);
  }

  return status;
}
