/*
** ADM-XRC Series Driver
**
** (C) Copright 2001-2005 Alpha Data
**
** methods.c - Model-specific routines.
**
** ----------------------------------------------------------------------------
*/

#include "drvparam.h"
#include "flash.h"
#include "methods.h"

/*
** Receive up to 32 bits from a 93x6 EEPROM
*/
static void
receive93x6Bits(
  OS_BAR_HANDLE handle,
  PLX9080REGS* plx,
  os_uint32* bits,
  unsigned int len,
  os_uint32* cntrl)
{
  os_uint32 val32 = 0;
  os_uint32 readval;
  unsigned int i;
  
  for (i = 0; i < len; i++) {
    val32 <<= 1;
    os_write_register_le32(handle, &plx->CNTRL, *cntrl);
    os_read_register_le32(handle, &plx->CNTRL);
    os_delay_busy_us(1);
    os_write_register_le32(handle, &plx->CNTRL, *cntrl | 0x01000000U);
    os_read_register_le32(handle, &plx->CNTRL);
    os_delay_busy_us(1);
    os_write_register_le32(handle, &plx->CNTRL, *cntrl);
    readval = os_read_register_le32(handle, &plx->CNTRL);
    os_delay_busy_us(1);
    if (readval & 0x08000000) {
      val32 |= 0x1U;
    }
  }
  
  *bits = val32;
}

/*
** Send up to 32 bits to a 93x6 EEPROM
**
** With PCI9656, always leaves EEDO pin disabled on exit.
*/
static void
send93x6Bits(
  OS_BAR_HANDLE handle,
  PLX9080REGS* plx,
  os_boolean pci9656,
  os_uint32 bits,
  unsigned int len,
  os_uint32* cntrl)
{
  unsigned int	i;

  if (pci9656) {
    /* Enable driver on EEDO pin */
    *cntrl &= ~0x80000000U;
  }
  os_write_register_le32(handle, &plx->CNTRL, *cntrl);
  os_read_register_le32(handle, &plx->CNTRL);
  os_delay_busy_us(1);
  
  /* Send instruction to EEPROM one bit at a time */
  for (i = 0; i < len; i++) {
    /* Pre-clear the EEPROM data bit */
    *cntrl &= ~0x04000000U;
    /* Set the EEPROM data bit if necessary */
    *cntrl |= (bits & (0x1U << (len - 1))) ? 0x04000000: 0x00000000;

    /* Clock the data bit to the EEPROM */
    os_write_register_le32(handle, &plx->CNTRL, *cntrl);
    os_read_register_le32(handle, &plx->CNTRL);
    os_delay_busy_us(1);
    os_write_register_le32(handle, &plx->CNTRL, *cntrl | 0x01000000);
    os_read_register_le32(handle, &plx->CNTRL);
    os_delay_busy_us(1);
    os_write_register_le32(handle, &plx->CNTRL, *cntrl);
    os_read_register_le32(handle, &plx->CNTRL);
    os_delay_busy_us(1);
    
    bits <<= 1;
  }
  
  if (pci9656) {
    /* Disable driver on EEDO pin */
    *cntrl |= 0x80000000U;
    os_write_register_le32(handle, &plx->CNTRL, *cntrl);
    os_read_register_le32(handle, &plx->CNTRL);
    os_delay_busy_us(1);    
  }
}

/*
** Send a command to a 93x6 EEPROM
*/
static void
send93x6Command(
  OS_BAR_HANDLE handle,
  PLX9080REGS* plx,
  os_boolean pci9656,
  os_uint32 instr,
  unsigned int len,
  os_uint32* cntrl)
{	
  /* Chip select asserted */
  *cntrl |=  0x02000000U;
  /* Clear instruction and clock bits */
  *cntrl &= ~0x05000000U;
  os_write_register_le32(handle, &plx->CNTRL, *cntrl);
  os_read_register_le32(handle, &plx->CNTRL);
  os_delay_busy_us(1);

  /*
  ** Toggle EEPROM's chip select (to get it out of Shift Register Mode)
  ** and assert the chip select.
  */
  os_write_register_le32(handle, &plx->CNTRL, *cntrl & ~0x02000000U);
  os_read_register_le32(handle, &plx->CNTRL);
  os_delay_busy_us(1);
  os_write_register_le32(handle, &plx->CNTRL, *cntrl);
  os_read_register_le32(handle, &plx->CNTRL);
  os_delay_busy_us(1);
  
  /* Send the instruction to the EEPROM */
  send93x6Bits(handle, plx, pci9656, instr, len, cntrl);
}

/*
** -----------------------------------------------------------------------
** The error method; should never be called unless something goes wrong
** -----------------------------------------------------------------------
*/
MethodResult
methodError(
  DeviceContext* pDevCtx,
  ...)
{
  os_dbgprint(0, ("*** methodError: called, pDevCtx=0x%p\n", pDevCtx));

  return MethodResult_GeneralFailure;
}

/*
** -----------------------------------------------------------------------
** MethodConfigMem
** -----------------------------------------------------------------------
*/

MethodResult
methodConfigMemPlxbased(
  DeviceContext* pDevCtx,
  os_boolean     write,
  unsigned long  location,
  os_uint32*     data)
{
  MethodResult result = MethodResult_Success;
  PLX9080REGS* plx = pDevCtx->hw.adapter.pci9080;
  OS_BAR_HANDLE handle = pDevCtx->hw.adapterHandle;
  os_uint32    val32;
  os_uint32    cntrl;
  unsigned int j;
  unsigned int instr_len, addr_len;
  os_uint16    addr_mask;
  os_boolean   pci9656;
  
  switch (pDevCtx->devId.pci.subDevice) {
  case PCI_SUBDEVICE_ADATA_ADMXRC:
  case PCI_SUBDEVICE_ADATA_ADMXRC2L:
    addr_mask = 0x3fU;
    addr_len = 6;
    instr_len = 9;
    break;

  case PCI_SUBDEVICE_ADATA_ADMXRC2:
  case PCI_SUBDEVICE_ADATA_ADPWRC2:
  case PCI_SUBDEVICE_ADATA_ADPDRC2:
  case PCI_SUBDEVICE_ADATA_ADMXRC4LS:
  case PCI_SUBDEVICE_ADATA_ADMXRC4LX:
  case PCI_SUBDEVICE_ADATA_ADMXRC4SX:
    addr_mask = 0xffU;
    addr_len = 8;
    instr_len = 11;
    break;

  default:
    os_assert(FALSE);
    return MethodResult_GeneralFailure;
  }
  
  if (location > (0x1U << addr_len)) {
    return MethodResult_InvalidLocation;
  }
  
  pci9656 = (pDevCtx->devId.pci.subDevice == PCI_SUBDEVICE_ADATA_ADMXRC2 ||
	     pDevCtx->devId.pci.subDevice == PCI_SUBDEVICE_ADATA_ADPWRC2 ||
	     pDevCtx->devId.pci.subDevice == PCI_SUBDEVICE_ADATA_ADPDRC2 ||
	     pDevCtx->devId.pci.subDevice == PCI_SUBDEVICE_ADATA_ADMXRC4LS ||
	     pDevCtx->devId.pci.subDevice == PCI_SUBDEVICE_ADATA_ADMXRC4LX ||
	     pDevCtx->devId.pci.subDevice == PCI_SUBDEVICE_ADATA_ADMXRC4SX);
  
  /*
  ** FIXME: this should use pDevCtx->shadow.regs.plxbased.cntrl and be
  ** spinlocked appropriately.
  */
  cntrl = os_read_register_le32(handle, &plx->CNTRL);
  
  if (write) {
    val32 = *data;
    
    /* Enable writing to EEPROM */
    send93x6Command(handle, plx, pci9656, EE_WREN << (addr_len - 2), instr_len, &cntrl);

    /* Send write command to EEPROM */
    send93x6Command(handle, plx, pci9656, (EE_WRITE << addr_len) | (location & addr_mask), instr_len, &cntrl);

    /* Send data word to EEPROM */
    send93x6Bits(handle, plx, pci9656, val32, 16, &cntrl);
		
    /* Activate EEPROM's status mechanism: deselect and reselect */
    os_write_register_le32(handle, &plx->CNTRL, cntrl & ~0x02000000);
    os_read_register_le32(handle, &plx->CNTRL);
    os_delay_busy_us(1);
    os_write_register_le32(handle, &plx->CNTRL, cntrl);
    os_read_register_le32(handle, &plx->CNTRL);
    os_delay_busy_us(1);

    /*
    ** Poll chip status until write cycle completes (or times out)
    */
    for (j = 0; j < 20000; j++) {
      if (os_read_register_le32(handle, &plx->CNTRL) & 0x08000000U) {
	result = MethodResult_Timeout;
      }
    }
    
    /* Disable writing to EEPROM */
    send93x6Command(handle, plx, pci9656, EE_WDS << addr_len, instr_len, &cntrl);
  } else {
    /* Send read command to EEPROM */
    send93x6Command(handle, plx, pci9656, (EE_READ << addr_len) | (location & addr_mask), instr_len, &cntrl);
		
    /* Get 16-bit word from EEPROM */
    receive93x6Bits(handle, plx, &val32, 16, &cntrl);
    
    *data = (os_uint16)val32;
  }
  
  return result;
}

MethodResult
methodConfigMemXpl(
  DeviceContext* pDevCtx,
  os_boolean     bWrite,
  unsigned long  location,
  os_uint32*     data)
{
  MethodResult	result;
  os_uint32 tmp;

  if (location >= 0x40000U) {
    return MethodResult_InvalidLocation;
  }
  
  if (bWrite) {
    tmp = os_cpu_to_le32(*data);
    result = writeFlash(pDevCtx, location * sizeof(os_uint32), sizeof(os_uint32), &tmp);
  } else {
    result = readFlash(pDevCtx, location * sizeof(os_uint32), sizeof(os_uint32), &tmp);
    *data = os_le32_to_cpu(tmp);
  }
  
  return result;
}

MethodResult
methodConfigMemAmc5a2(
  DeviceContext* pDevCtx,
  os_boolean bWrite,
  unsigned long location,
  os_uint32* data)
{
  MethodResult	result;
  os_uint32 tmp;
  
  if (location >= 0x4000U) {
    return MethodResult_InvalidLocation;
  }
  
  if (bWrite) {
    tmp = os_cpu_to_le32(*data);
    result = writeFlash(pDevCtx, VPD_ADDR_ADMAMC5A2 + location * sizeof(os_uint32), sizeof(os_uint32), &tmp);
  } else {
    result = readFlash(pDevCtx, VPD_ADDR_ADMAMC5A2 + location * sizeof(os_uint32), sizeof(os_uint32), &tmp);
    *data = os_le32_to_cpu(tmp);
  }
  
  return result;
}

MethodResult
methodConfigMemBbp(
  DeviceContext* pDevCtx,
  os_boolean     bWrite,
  unsigned long  location,
  os_uint32*     data)
{
  MethodResult	result;
  os_uint32 tmp;

  if (location >= 0x40000U) {
    return MethodResult_InvalidLocation;
  }
  
  if (bWrite) {
    tmp = os_cpu_to_le32(*data);
    result = writeFlash(pDevCtx, VPD_ADDR_ADCBBP + location * sizeof(os_uint32), sizeof(os_uint32), &tmp);
  } else {
    result = readFlash(pDevCtx, VPD_ADDR_ADCBBP + location * sizeof(os_uint32), sizeof(os_uint32), &tmp);
    *data = os_le32_to_cpu(tmp);
  }
  
  return result;
}

MethodResult
methodConfigMemXrc5t1VpdBram(
  DeviceContext* pDevCtx,
  os_boolean     bWrite,
  unsigned long  location,
  os_uint32*     data)
{
  MethodResult result = MethodResult_Success;
  os_uint32 val32, flpage, flctl;
  XRC4FXSPACE1REGS* pRegs;
  OS_BAR_HANDLE handle;

  if (location >= 0x4000U) {
    return MethodResult_InvalidLocation;
  }

  pRegs = pDevCtx->hw.control.xrc4fx;
  handle = pDevCtx->hw.controlHandle;

  flpage = location & 0x3fe0000U;
  os_write_register_le32(handle, &pRegs->FLPAGE, flpage);

  if (bWrite) {
    return MethodResult_GeneralFailure;
  } else {
    /* Set BlockRAM address for low 16 bits */
    flctl = (location << 2U) & 0x1fffeU;
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);

    /* Read low 16 bits */
    val32 = (os_uint16)os_read_register_le32(handle, &pRegs->FLDATA);

    /* Set BlockRAM address for high 16 bits */
    flctl = ((location << 2U) | 0x2U) & 0x1fffeU;
    os_write_register_le32(handle, &pRegs->FLCTL, flctl);

    /* Read high 16 bits */
    val32 |= ((os_uint16)os_read_register_le32(handle, &pRegs->FLDATA)) << 16U;

    os_dbgprint(9, ("methodConfigMemXrc5t1VpdBram: val32=0x%lx\n", (unsigned long)val32));

    *data = val32;
  }

  return result;
}

/*
** -----------------------------------------------------------------------
** MethodConfigSpace
** -----------------------------------------------------------------------
*/

MethodResult
methodConfigSpacePlxbased(
  DeviceContext* pDevCtx,
  os_boolean     write,
  unsigned int   space,
  os_uint32*     value)
{
  PLX9080REGS* pRegs = pDevCtx->hw.adapter.pci9080;
  OS_BAR_HANDLE handle = pDevCtx->hw.adapterHandle;
  os_uint32 lbrd0;
  os_uint32 f;

  if (space > 1) {
    return MethodResult_NotSupported;
  }
  
  if (write) {
    f = *value;
    lbrd0 = os_read_register_le32(handle, &pDevCtx->hw.adapter.pci9080->LBRD0);
    
    if (f & ADMXRC2_SPACE_SET_WIDTH) {
      if ((f & ADMXRC2_SPACE_MASK_WIDTH) == ADMXRC2_SPACE_WIDTH_DEFAULT) {
	f = (f & ~ADMXRC2_SPACE_MASK_WIDTH) | ADMXRC2_SPACE_WIDTH_32;
      }
    } else {
      f = (f & ~ADMXRC2_SPACE_MASK_WIDTH) | (pDevCtx->info.spaceConfig[space] & ADMXRC2_SPACE_MASK_WIDTH);
    }
    
    if (f & ADMXRC2_SPACE_SET_PREFETCH) {
      if ((f & ADMXRC2_SPACE_MASK_PREFETCH) == ADMXRC2_SPACE_PREFETCH_DEFAULT) {
	f = (f & ~ADMXRC2_SPACE_MASK_PREFETCH) | ADMXRC2_SPACE_PREFETCH_MAXIMUM;
      }
    } else {
      f = (f & ~ADMXRC2_SPACE_MASK_PREFETCH) | (pDevCtx->info.spaceConfig[space] & ADMXRC2_SPACE_MASK_PREFETCH);
    }
    
    if (f & ADMXRC2_SPACE_SET_BURST) {
      if ((f & ADMXRC2_SPACE_MASK_BURST) == ADMXRC2_SPACE_BURST_DEFAULT) {
	f = (f & ~ADMXRC2_SPACE_MASK_BURST) | ADMXRC2_SPACE_BURST_ENABLED;
      }
    } else {
      f = (f & ~ADMXRC2_SPACE_MASK_BURST) | (pDevCtx->info.spaceConfig[space] & ADMXRC2_SPACE_MASK_BURST);
    }
    
    f &= ~(ADMXRC2_SPACE_SET_WIDTH | ADMXRC2_SPACE_SET_BURST | ADMXRC2_SPACE_SET_PREFETCH);
    
    switch (f & ADMXRC2_SPACE_MASK_WIDTH) {
    case ADMXRC2_SPACE_WIDTH_8:
      lbrd0 = (lbrd0 & ~0x3U) | 0x0U;
      break;
      
    case ADMXRC2_SPACE_WIDTH_16:
      lbrd0 = (lbrd0 & ~0x3U) | 0x1U;
      break;
      
    case ADMXRC2_SPACE_WIDTH_32:
      lbrd0 = (lbrd0 & ~0x3U) | 0x2U;
      break;
      
    case ADMXRC2_SPACE_WIDTH_64:
      return MethodResult_NotSupported;
      
    default:
      return MethodResult_InvalidParameter;
    }
    
    switch (f & ADMXRC2_SPACE_MASK_PREFETCH) {
    case ADMXRC2_SPACE_PREFETCH_MINIMUM:
      /* No prefetch */
      lbrd0 = (lbrd0 & ~0x100U) | (0x1U << 8);
      break;
      
    case ADMXRC2_SPACE_PREFETCH_NORMAL:
      /* 64 bytes prefetch */
      lbrd0 = (lbrd0 & ~0x500U) | (0x1U << 10);
      lbrd0 = (lbrd0 & ~0x7800U) | (0x0U << 11);
      break;
      
    case ADMXRC2_SPACE_PREFETCH_MAXIMUM:
      /* Unlimited prefetch */
      lbrd0 = (lbrd0 & ~0x500U) | (0x0U << 10);
      lbrd0 = (lbrd0 & ~0x7800U) | (0x0U << 11);
      break;
      
    default:
      return MethodResult_InvalidParameter;
    }
    
    switch (f & ADMXRC2_SPACE_MASK_BURST) {
    case ADMXRC2_SPACE_BURST_DISABLED:
      lbrd0 = (lbrd0 & ~0x1000000U) | (0x0U << 24);
      break;
      
    case ADMXRC2_SPACE_BURST_ENABLED:
      lbrd0 = (lbrd0 & ~0x1000000U) | (0x1U << 24);
      break;
      
    default:
      return MethodResult_InvalidParameter;
    }
    
    os_write_register_le32(handle, &pRegs->LBRD0, lbrd0);
    pDevCtx->info.spaceConfig[space] = f;
  } else {
    *value = pDevCtx->info.spaceConfig[space];
  }
  
  return MethodResult_Success;
}

MethodResult
methodConfigSpaceXplrev0(
  DeviceContext* pDevCtx,
  os_boolean write,
  unsigned int space,
  os_uint32* value)
{
  XPLBRIDGEREGS* pRegs = pDevCtx->hw.adapter.xpl;
  OS_BAR_HANDLE handle = pDevCtx->hw.adapterHandle;
  os_uint32 ls0cfg;
  os_uint32 f = *value;

  if (space > 1) {
    return MethodResult_NotSupported;
  }

  if (write) {
    ls0cfg = os_read_register_le32(handle, &pDevCtx->hw.adapter.xpl->LS0CFG);
		
    if (f & ADMXRC2_SPACE_SET_WIDTH) {
      if ((f & ADMXRC2_SPACE_MASK_WIDTH) == ADMXRC2_SPACE_WIDTH_DEFAULT) {
	f = (f & ~ADMXRC2_SPACE_MASK_WIDTH) | ADMXRC2_SPACE_WIDTH_32;
      }
    } else {
      f = (f & ~ADMXRC2_SPACE_MASK_WIDTH) | (pDevCtx->info.spaceConfig[space] & ADMXRC2_SPACE_MASK_WIDTH);
    }

    if (f & ADMXRC2_SPACE_SET_PREFETCH) {
      if ((f & ADMXRC2_SPACE_MASK_PREFETCH) == ADMXRC2_SPACE_PREFETCH_DEFAULT) {
	f = (f & ~ADMXRC2_SPACE_MASK_PREFETCH) | ADMXRC2_SPACE_PREFETCH_NORMAL;
      }
    } else {
      f = (f & ~ADMXRC2_SPACE_MASK_PREFETCH) | (pDevCtx->info.spaceConfig[space] & ADMXRC2_SPACE_MASK_PREFETCH);
    }

    if (f & ADMXRC2_SPACE_SET_BURST) {
      if ((f & ADMXRC2_SPACE_MASK_BURST) == ADMXRC2_SPACE_BURST_DEFAULT) {
	f = (f & ~ADMXRC2_SPACE_MASK_BURST) | ADMXRC2_SPACE_BURST_ENABLED;
      }
    } else {
      f = (f & ~ADMXRC2_SPACE_MASK_BURST) | (pDevCtx->info.spaceConfig[space] & ADMXRC2_SPACE_MASK_BURST);
    }

    f &= ~(ADMXRC2_SPACE_SET_WIDTH | ADMXRC2_SPACE_SET_BURST | ADMXRC2_SPACE_SET_PREFETCH);

    switch (f & ADMXRC2_SPACE_MASK_WIDTH) {
    case ADMXRC2_SPACE_WIDTH_8:
    case ADMXRC2_SPACE_WIDTH_16:
      return MethodResult_NotSupported;
			
    case ADMXRC2_SPACE_WIDTH_32:
      ls0cfg = (ls0cfg & ~0x3U) | 0x2U;
      break;
			
    case ADMXRC2_SPACE_WIDTH_64:
      ls0cfg = (ls0cfg & ~0x3U) | 0x3U;
      break;
			
    default:
      return MethodResult_InvalidParameter;
    }
		
    switch (f & ADMXRC2_SPACE_MASK_PREFETCH) {
    case ADMXRC2_SPACE_PREFETCH_MINIMUM:
      /* 8 bytes prefetch */
      ls0cfg = (ls0cfg & ~0x3cU) | (0x0U << 5);
      break;
			
    case ADMXRC2_SPACE_PREFETCH_NORMAL:
      /* 64 bytes prefetch */
      ls0cfg = (ls0cfg & ~0x3cU) | (0x1U << 5) | (0x2U << 2);
      break;
			
    case ADMXRC2_SPACE_PREFETCH_MAXIMUM:
      /* Unlimited prefetch */
      ls0cfg = (ls0cfg & ~0x3cU) | (0x1U << 5) | (0x4U << 2);
      break;
			
    default:
      return MethodResult_InvalidParameter;
    }
		
    switch (f & ADMXRC2_SPACE_MASK_BURST) {
    case ADMXRC2_SPACE_BURST_DISABLED:
      return MethodResult_NotSupported;
			
    case ADMXRC2_SPACE_BURST_ENABLED:
      break;
			
    default:
      return MethodResult_InvalidParameter;
    }
		
    os_write_register_le32(handle, &pRegs->LS0CFG, ls0cfg);
    pDevCtx->info.spaceConfig[space] = f;
  } else {
    *value = pDevCtx->info.spaceConfig[space];
  }

  return MethodResult_Success;
}

MethodResult
methodConfigSpaceXplrev1(
  DeviceContext* pDevCtx,
  os_boolean write,
  unsigned int space,
  os_uint32* value)
{
  XPLBRIDGEREGS* pRegs = pDevCtx->hw.adapter.xpl;
  OS_BAR_HANDLE handle = pDevCtx->hw.adapterHandle;
  os_uint32 ls0cfg;
  os_uint32 f = *value;

  if (space > 1) {
    return MethodResult_NotSupported;
  }

  if (write) {
    ls0cfg = os_read_register_le32(handle, &pRegs->LS0CFG);
		
    if (f & ADMXRC2_SPACE_SET_WIDTH) {
      if ((f & ADMXRC2_SPACE_MASK_WIDTH) == ADMXRC2_SPACE_WIDTH_DEFAULT) {
	f = (f & ~ADMXRC2_SPACE_MASK_WIDTH) | ADMXRC2_SPACE_WIDTH_32;
      }
    } else {
      f = (f & ~ADMXRC2_SPACE_MASK_WIDTH) | (pDevCtx->info.spaceConfig[space] & ADMXRC2_SPACE_MASK_WIDTH);
    }

    if (f & ADMXRC2_SPACE_SET_PREFETCH) {
      if ((f & ADMXRC2_SPACE_MASK_PREFETCH) == ADMXRC2_SPACE_PREFETCH_DEFAULT) {
	f = (f & ~ADMXRC2_SPACE_MASK_PREFETCH) | ADMXRC2_SPACE_PREFETCH_NORMAL;
      }
    } else {
      f = (f & ~ADMXRC2_SPACE_MASK_PREFETCH) | (pDevCtx->info.spaceConfig[space] & ADMXRC2_SPACE_MASK_PREFETCH);
    }

    if (f & ADMXRC2_SPACE_SET_BURST) {
      if ((f & ADMXRC2_SPACE_MASK_BURST) == ADMXRC2_SPACE_BURST_DEFAULT) {
	f = (f & ~ADMXRC2_SPACE_MASK_BURST) | ADMXRC2_SPACE_BURST_ENABLED;
      }
    } else {
      f = (f & ~ADMXRC2_SPACE_MASK_BURST) | (pDevCtx->info.spaceConfig[space] & ADMXRC2_SPACE_MASK_BURST);
    }

    f &= ~(ADMXRC2_SPACE_SET_WIDTH | ADMXRC2_SPACE_SET_BURST | ADMXRC2_SPACE_SET_PREFETCH);

    switch (f & ADMXRC2_SPACE_MASK_WIDTH) {
    case ADMXRC2_SPACE_WIDTH_8:
    case ADMXRC2_SPACE_WIDTH_16:
      return MethodResult_NotSupported;
			
    case ADMXRC2_SPACE_WIDTH_32:
      ls0cfg = (ls0cfg & ~0x3U) | 0x2U;
      break;
			
    case ADMXRC2_SPACE_WIDTH_64:
      ls0cfg = (ls0cfg & ~0x3U) | 0x3U;
      break;
			
    default:
      return MethodResult_InvalidParameter;
    }
		
    switch (f & ADMXRC2_SPACE_MASK_PREFETCH) {
    case ADMXRC2_SPACE_PREFETCH_MINIMUM:
      /* No prefetch */
      ls0cfg = (ls0cfg & ~0x3cU) | (0x0U << 5);
      break;
			
    case ADMXRC2_SPACE_PREFETCH_NORMAL:
      /* 64 bytes prefetch */
      ls0cfg = (ls0cfg & ~0x3cU) | (0x1U << 5) | (0x2U << 2);
      break;
			
    case ADMXRC2_SPACE_PREFETCH_MAXIMUM:
      /* Unlimited prefetch */
      ls0cfg = (ls0cfg & ~0x3cU) | (0x1U << 5) | (0x4U << 2);
      break;
			
    default:
      return MethodResult_InvalidParameter;
    }
		
    switch (f & ADMXRC2_SPACE_MASK_BURST) {
    case ADMXRC2_SPACE_BURST_DISABLED:
      ls0cfg &= ~(0x1U << 6);
      break;
			
    case ADMXRC2_SPACE_BURST_ENABLED:
      ls0cfg |= (0x1U << 6);
      break;
			
    default:
      return MethodResult_InvalidParameter;
    }
		
    os_write_register_le32(handle, &pRegs->LS0CFG, ls0cfg);
    pDevCtx->info.spaceConfig[space] = f;
  } else {
    *value = pDevCtx->info.spaceConfig[space];
  }

  return MethodResult_Success;
}
