/*
** ADM-XRC Series Driver
**
** (C) Copright 2001-2008 Alpha Data
**
** methods_hwinit.c - Model-specific routines for initializing card registers.
*/

#include "drvparam.h"
#include "device.h"
#include "methods.h"

static void
initCard9080(
  DeviceContext* pDevCtx)
{
  PLX9080REGS* pRegs = pDevCtx->hw.adapter.pci9080;
  OS_BAR_HANDLE handle = pDevCtx->hw.adapterHandle;

  os_dbgprint(1, ("initCard9080: entered\n"));

  /* Enable Direct Master access */
  os_write_register_le32(handle, &pRegs->DMRR,   0xFF800000U);
  os_write_register_le32(handle, &pRegs->DMLBAM, 0xFF000000U);
  os_write_register_le32(handle, &pRegs->DMLBAI, 0x0U);
  os_write_register_le32(handle, &pRegs->DMPBAM, 0x1U);

  if (admxrcDriverParams.pciReadMultiple) {
    /* Use PCI read multiple commands when a bus master */
    os_dbgprint(1, ("initCard9080: selecting PCI read multiple command for bus master reads\n"));
    pDevCtx->shadow.regs.plxbased.cntrl = os_read_register_le32(handle, &pRegs->CNTRL);
    pDevCtx->shadow.regs.plxbased.cntrl &= ~0xf0fU;
    pDevCtx->shadow.regs.plxbased.cntrl |= 0xc0cU;
    os_write_register_le32(handle, &pRegs->CNTRL, pDevCtx->shadow.regs.plxbased.cntrl);
  }

  /* Establish the DMACSR register: both DMA channels initially enabled */
  pDevCtx->shadow.regs.plxbased.dmacsr = 0x0101U;
  os_write_register_le16(handle, &pRegs->DMACSR.asWord, pDevCtx->shadow.regs.plxbased.dmacsr); 
}

static void
initCard9656(
  DeviceContext* pDevCtx)
{
  PLX9656REGS* pRegs = pDevCtx->hw.adapter.pci9656;
  OS_BAR_HANDLE handle = pDevCtx->hw.adapterHandle;

  os_dbgprint(1, ("initCard9656: entered\n"));

  /*
  ** Currently, card initialization for the PCI9656 is the same as for the
  ** PCI9080, but with a few extra bits and pieces.
  */
  initCard9080(pDevCtx);

  if (ADMXRC_9656_ERRATA_WORKAROUND) {
    os_uint32 val;

    os_dbgprint(1, ("initCard9656: using PCI9656 rev AC/AD workarounds\n"));

    /*
    ** Implement workarounds for PCI9656 rev AC and AD errata
    */
    
    /* DMPBAM register */
    val = os_read_register_le32(handle, &pRegs->DMPBAM);
    val &= ~(0x1U <<  2); /* Disable Direct Master read-ahead */
    /* Do not prefetch more than 8 QWORDs for a Direct Master read */
    val &= ~(0x1U <<  3);
    val |=  (0x1U << 12);
    os_write_register_le32(handle, &pRegs->DMPBAM, val);
    
    /* EROMBA register */
    val = os_read_register_le32(handle, &pRegs->EROMBA);
    val &= ~(0x1U <<  4); /* Don't use local bus backoff for Direct Master transfers */
    os_write_register_le32(handle, &pRegs->EROMBA, val);
    
    /* LBRD0 register */
    val = os_read_register_le32(handle, &pRegs->LBRD0);
    val |=  (0x1U <<  8); /* Prefetching permitted for Direct Slave Space 0 reads */
    os_write_register_le32(handle, &pRegs->LBRD0, val);
    
    /* LBRD1 register */
    val = os_read_register_le32(handle, &pRegs->LBRD1);
    val |=  (0x1U <<  8); /* Prefetching permitted for Direct Slave Space 1 reads */
    os_write_register_le32(handle, &pRegs->LBRD1, val);
    
    /* LMISC2 register */
    val = os_read_register_le8(handle, &pRegs->LMISC2);
    val |=  (0x1U <<  5); /* Reserve 1 location in Direct Slave write FIFO */
    os_write_register_le8(handle, &pRegs->LMISC2, val);
    
    /* MARBR register */
    val = os_read_register_le32(handle, &pRegs->MARBR);
    val &= ~(0x1U << 16); /* Disable local bus latency timer */
    val &= ~(0x1U << 23); /* Assert REQ# on PCI for entire DMA transfer */
    val |=  (0x1U << 24); /* Operate in PCI 2.1 mode */
    val |=  (0x1U << 26); /* Flush Direct Slave read FIFO after Direct Slave write */
    val &= ~(0x1U << 28); /* Direct slave read ahead mode disabled */
    os_write_register_le32(handle, &pRegs->MARBR, val);
  }
}

/*
** -----------------------------------------------------------------------
** MethodHwInit
** -----------------------------------------------------------------------
*/

MethodResult
methodHwInitPci9080based(
  DeviceContext* pDevCtx)
{
  os_dbgprint(2, ("methodHwInitPci9080based: entered\n"));

  initCard9080(pDevCtx);

  return MethodResult_Success;
}

MethodResult
methodHwInitPci9656based(
  DeviceContext* pDevCtx)
{
  os_dbgprint(2, ("methodHwInitPci9656based: entered\n"));

  initCard9656(pDevCtx);

  return MethodResult_Success;
}

MethodResult
methodHwInitXplrev0(
  DeviceContext* pDevCtx)
{
  os_dbgprint(2, ("methodHwInitXplrev0: entered\n"));

  /* Nothing to do */
  return MethodResult_Success;
}

MethodResult
methodHwInitXplrev1(
  DeviceContext* pDevCtx)
{
  XPLBRIDGEREGS* pRegs = pDevCtx->hw.adapter.xpl;
  OS_BAR_HANDLE handle = pDevCtx->hw.adapterHandle;
  unsigned int i;

  os_dbgprint(2, ("methodHwInitXplrev1: entered\n"));

  /* Initialise DMA channel state */
  for (i = 0; i < NUM_XPLBRIDGE_DMA_CHANNEL; i++) {
    pDevCtx->shadow.regs.xpl.dxctl[i] = 0x4U;
    os_write_register_le32(handle, &pRegs->DMA[i].CTL, pDevCtx->shadow.regs.xpl.dxctl[i]);
  }

  return MethodResult_Success;
}

MethodResult
methodHwInitXp(
  DeviceContext* pDevCtx)
{
  XPLBRIDGEREGS* pAdapterRegs = pDevCtx->hw.adapter.xpl;
  OS_BAR_HANDLE adapterHandle = pDevCtx->hw.adapterHandle;
  XPSPACE1REGS* pControlRegs = pDevCtx->hw.control.xp;
  OS_BAR_HANDLE controlHandle = pDevCtx->hw.controlHandle;
  unsigned int i;
  
  os_dbgprint(2, ("methodHwInitXp: entered\n"));

  /* Initialise DMA channel state */
  for (i = 0; i < NUM_XPLBRIDGE_DMA_CHANNEL; i++) {
    pDevCtx->shadow.regs.xp.dxctl[i] = 0x4U;
    os_write_register_le32(adapterHandle, &pAdapterRegs->DMA[i].CTL, pDevCtx->shadow.regs.xp.dxctl[i]);
    os_read_register_le32(adapterHandle, &pAdapterRegs->DMA[i].CTL);
  }
    
  /* Initialise Flash access stuff */
  pDevCtx->shadow.regs.xp.fldata = (0x3U << 26);
  os_write_register_le32(controlHandle, &pControlRegs->FLDATA, pDevCtx->shadow.regs.xp.fldata);
  os_read_register_le32(controlHandle, &pControlRegs->FLDATA);
    
  return MethodResult_Success;
}

MethodResult
methodHwInitXrc4ls(
  DeviceContext* pDevCtx)
{
  PLX9656REGS* pAdapterRegs = pDevCtx->hw.adapter.pci9656;
  OS_BAR_HANDLE adapterHandle = pDevCtx->hw.adapterHandle;
  CONTROLXRC4LS* pControlRegs = pDevCtx->hw.control.xrc4ls;
  OS_BAR_HANDLE controlHandle = pDevCtx->hw.controlHandle;

  os_dbgprint(2, ("methodHwInitXrc4ls: entered\n"));

  if (pDevCtx->dma.info.numChan) {
    /* Establish the DMACSR register: both DMA channels initially enabled */
    pDevCtx->shadow.regs.xrc4ls.dmacsr = 0x0101U;
    os_write_register_le16(adapterHandle, &pAdapterRegs->DMACSR.asWord, (os_uint16)pDevCtx->shadow.regs.xrc4ls.dmacsr);
  }

  /* Initialise Flash access stuff */
  pDevCtx->shadow.regs.xrc4ls.flctl = (0x1U << 26);
  os_write_register_le32(controlHandle, &pControlRegs->FLCTL, pDevCtx->shadow.regs.xrc4ls.flctl);
  os_read_register_le32(controlHandle, &pControlRegs->FLCTL);
    
  /* Initialise FPGA SelectMap mode */
  pDevCtx->shadow.regs.xrc4ls.fpctl = 0;
  if (admxrcDriverParams.model.xrc4ls.smap32Bit) {
    pDevCtx->shadow.regs.xrc4ls.fpctl |= (0x1U << 5);
  }
  os_write_register_le32(controlHandle, &pControlRegs->FPCTL, pDevCtx->shadow.regs.xrc4ls.fpctl);
  os_read_register_le32(controlHandle, &pControlRegs->FPCTL);
    
  return MethodResult_Success;
}

MethodResult
methodHwInitXrc4fx(
  DeviceContext* pDevCtx)
{
  XRC4FXBRIDGEREGS* pAdapterRegs = pDevCtx->hw.adapter.xrc4fx;
  OS_BAR_HANDLE adapterHandle = pDevCtx->hw.adapterHandle;
  XRC4FXSPACE1REGS* pControlRegs = pDevCtx->hw.control.xrc4fx;
  OS_BAR_HANDLE controlHandle = pDevCtx->hw.controlHandle;
  unsigned int i;
  
  os_dbgprint(2, ("methodHwInitXrc4fx: entered\n"));

  /* Initialise DMA channel state */
  for (i = 0; i < NUM_XRC4FXBRIDGE_DMA_CHANNEL; i++) {
    pDevCtx->shadow.regs.xrc4fx.dxctl[i] = 0x4U;
    os_write_register_le32(adapterHandle, &pAdapterRegs->DMA[i].CTL, pDevCtx->shadow.regs.xrc4fx.dxctl[i]);
    os_read_register_le32(adapterHandle, &pAdapterRegs->DMA[i].CTL);
  }
    
  /* Initialise Flash access stuff */
  pDevCtx->shadow.regs.xrc4fx.flctl = 0;
  os_write_register_le32(controlHandle, &pControlRegs->FLCTL, pDevCtx->shadow.regs.xrc4fx.flctl);
  os_read_register_le32(controlHandle, &pControlRegs->FLCTL);
  
  /* Initialise FPGA SelectMap mode */
  pDevCtx->shadow.regs.xrc4fx.fpctl = 0;
  if (admxrcDriverParams.model.xrc4fx.smap32Bit) {
    pDevCtx->shadow.regs.xrc4fx.fpctl |= (0x1U << 5);
  }
  os_write_register_le32(controlHandle, &pControlRegs->FPCTL, pDevCtx->shadow.regs.xrc4fx.fpctl);
  os_read_register_le32(controlHandle, &pControlRegs->FPCTL);
    
  return MethodResult_Success;
}

MethodResult
methodHwInitXrce4fx(
  DeviceContext* pDevCtx)
{
  XPLBRIDGEREGS* pAdapterRegs = pDevCtx->hw.adapter.xpl;
  OS_BAR_HANDLE adapterHandle = pDevCtx->hw.adapterHandle;
  XRCE4FXSPACE1REGS* pControlRegs = pDevCtx->hw.control.xrce4fx;
  OS_BAR_HANDLE controlHandle = pDevCtx->hw.controlHandle;
  unsigned int i;
  
  os_dbgprint(1, ("methodHwInitXrce4fx: entered\n"));
  
  /* Initialise DMA channel state */
  for (i = 0; i < NUM_XPLBRIDGE_DMA_CHANNEL; i++) {
    pDevCtx->shadow.regs.xrce4fx.dxctl[i] = 0x4U;
    os_write_register_le32(adapterHandle, &pAdapterRegs->DMA[i].CTL, pDevCtx->shadow.regs.xrce4fx.dxctl[i]);
    os_read_register_le32(adapterHandle, &pAdapterRegs->DMA[i].CTL);
  }
  
  /* Initialise Flash access stuff */
  pDevCtx->shadow.regs.xrce4fx.flctl = 0;
  os_write_register_le32(controlHandle, &pControlRegs->FLCTL, pDevCtx->shadow.regs.xrce4fx.flctl);
  os_read_register_le32(controlHandle, &pControlRegs->FLCTL);
  
  /* Initialise FPGA SelectMap mode */
  pDevCtx->shadow.regs.xrce4fx.fpctl = 0;
  os_write_register_le32(controlHandle, &pControlRegs->FPCTL, pDevCtx->shadow.regs.xrce4fx.fpctl);
  os_read_register_le32(controlHandle, &pControlRegs->FPCTL);
  
  return MethodResult_Success;
}

MethodResult
methodHwInitBbp(
  DeviceContext* pDevCtx)
{
  XRC4FXBRIDGEREGS* pAdapterRegs = pDevCtx->hw.adapter.xrc4fx;
  OS_BAR_HANDLE adapterHandle = pDevCtx->hw.adapterHandle;
  BBPSPACE1REGS* pControlRegs = pDevCtx->hw.control.bbp;
  OS_BAR_HANDLE controlHandle = pDevCtx->hw.controlHandle;
  unsigned int i;
  
  os_dbgprint(2, ("methodHwInitBbp: entered\n"));

  /* Initialise DMA channel state */
  for (i = 0; i < NUM_XRC4FXBRIDGE_DMA_CHANNEL; i++) {
    pDevCtx->shadow.regs.bbp.dxctl[i] = 0x4U;
    os_write_register_le32(adapterHandle, &pAdapterRegs->DMA[i].CTL, pDevCtx->shadow.regs.bbp.dxctl[i]);
    os_read_register_le32(adapterHandle, &pAdapterRegs->DMA[i].CTL);
  }
    
  /* Initialise Flash access stuff */
  pDevCtx->shadow.regs.bbp.flctl = 0;
  os_write_register_le32(controlHandle, &pControlRegs->FLCTL, pDevCtx->shadow.regs.bbp.flctl);
  os_read_register_le32(controlHandle, &pControlRegs->FLCTL);
  
  /* Initialise FPGA SelectMap mode */
  pDevCtx->shadow.regs.bbp.fpctl[0] = 0;
  pDevCtx->shadow.regs.bbp.fpctl[1] = 0;
  if (admxrcDriverParams.model.bbp.smap32Bit) {
    pDevCtx->shadow.regs.bbp.fpctl[0] |= (0x1U << 5);
    pDevCtx->shadow.regs.bbp.fpctl[1] |= (0x1U << 5);
  }
  os_write_register_le32(controlHandle, &pControlRegs->FPCTL0, pDevCtx->shadow.regs.bbp.fpctl[0]);
  os_write_register_le32(controlHandle, &pControlRegs->FPCTL1, pDevCtx->shadow.regs.bbp.fpctl[1]);
  os_read_register_le32(controlHandle, &pControlRegs->FPCTL1);
    
  return MethodResult_Success;
}

