/*
** ADM-XRC Series Linux Driver
**
** (C) Copyright 2001-2004 Alpha Data
**
** platform-2.6.c - Platform abstraction layer routines for Linux 2.6.x
**
** This module contains routines that hide the underlying operating system
** and platform from the generic section of the driver, in order to make the
** driver more portable across various operating systems and platforms.
**
** ----------------------------------------------------------------------------
*/

#include "platform.h"

#define _os_page_is_high(k, i) PageHighMem((k)->maplist[i]);

int _os_dbglevel = 0; /* no debugging messages by default */

static int lcache_index = 0;

static void
os_init_dpc(
  OS_DPC* pDpc,
  OS_DEVICE_OBJECT* pDevObj,
  OS_DPC_FUNC* pHandler,
  void* pContext);

#if (LINUX_VERSION_CODE >= 0x020613)
static irqreturn_t
_os_isr(
  int             irq,
  void*           pContext)
#else
static int
_os_isr(
  int             irq,
  void*           pContext,
  struct pt_regs* regs)
#endif
{
#if OS_ATTACHED_IRQ_HAS_DPC
  OS_ATTACHED_IRQ* irqObj = (OS_ATTACHED_IRQ*)pContext;
  OS_ISR_STATUS isrStatus;

  isrStatus = irqObj->isr(irqObj->deviceContext);
  if (isrStatus == OS_ISR_SCHEDULE_DPC) {
    os_schedule_dpc(&irqObj->dpc, NULL);
  }
  return IRQ_RETVAL(isrStatus != OS_ISR_NOT_HANDLED);
#else
  OS_ATTACHED_IRQ* irqObj = (OS_ATTACHED_IRQ*)pContext;
  os_boolean bHandled;

  bHandled = irqObj->isr(irqObj->deviceContext);
  return IRQ_RETVAL(bHandled);
#endif
}

static void
_os_dpc_handler(
  unsigned long context)
{
  OS_DPC*           dpc = (OS_DPC*)context;  
  void*             arg;
  void*             dpcContext;
  OS_DPC_FUNC*      handler;

  handler = dpc->handler;
  dpcContext = dpc->context;
  arg = dpc->argument;

  handler(dpcContext, arg);
}

/*
** -------------------------------------------------------------------
** Exported to client code
** -------------------------------------------------------------------
*/

OS_STATUS
os_allow_mapping(
  OS_DEVICE_OBJECT*    pDevObj,
  os_uint64            phys,
  os_uint64            length,
  void*                context,
  OS_MAPPABLE_REGION** handle)
{
  OS_MAPPABLE_REGION* region;
  unsigned long startpage, endpage;

  os_dbgprint(3, ("os_allow_mapping: phys = 0x%8.8lx%8.8lx length = 0x%8.8lx%8.8lx context = 0x%p\n",
		  (unsigned long) os_uint64_hi(phys),
		  (unsigned long) os_uint64_lo(phys),
		  (unsigned long) os_uint64_hi(length),
		  (unsigned long) os_uint64_lo(length),
		  context));

  region = (OS_MAPPABLE_REGION*) os_kernel_malloc(sizeof(OS_MAPPABLE_REGION));
  if (region == NULL) {
    return OS_STATUS_NOMEMORY;
  }

  startpage = phys & os_page_mask_high;
  endpage = (phys + length + os_page_size - 1) & os_page_mask_high;

  region->pDevObj = pDevObj;
  region->pageAligned.phys = startpage;
  region->pageAligned.length = endpage - startpage;
  region->unaligned.phys = phys;
  region->unaligned.length = length;
  region->context = context;
  region->next = pDevObj->maps.regions;
  region->prev = &pDevObj->maps.regions;
  pDevObj->maps.regions = region;

  os_dbgprint(3, ("os_allow_mapping: created region 0x%p start 0x%lx-0x%lx\n",
		  region, (unsigned long) startpage, (unsigned long) endpage - 1));

  *handle = region;

  return OS_STATUS_SUCCESS;
}

void
os_complete_request(
  OS_IO_REQUEST* req)
{
  OS_DEVICE_OBJECT* pDevObj;
  OS_SPINLOCK_FLAGS f;

  os_dbgprint(4, ("os_complete_request: req=0x%p, status=0x%lx, length=%lu(0x%lx)\n",
		  (void*)req, (unsigned long)req->status.status, (unsigned long)req->status.length, (unsigned long)req->status.length));

  pDevObj = req->opaque.pDevObj;
  os_acquire_spinlock(&pDevObj->cancelLock, &f);
  req->opaque.completed = TRUE;
  wake_up(req->opaque.wait_q);
  os_release_spinlock(&pDevObj->cancelLock, f);
}

OS_STATUS
os_dma_mapper_create(
  OS_DEVICE_OBJECT* pDevObj,
  unsigned int maxEntry,
  OS_DMA_MAPPER** ppMapper)
{
  OS_STATUS status = OS_STATUS_SUCCESS;
  OS_DMA_MAPPER* pMapper = NULL;
  unsigned int i;

  pMapper = (OS_DMA_MAPPER*)os_kernel_malloc(sizeof(*pMapper) + maxEntry * sizeof(pMapper->opaque.bounce[0]));
  if (NULL == pMapper) {
    status = OS_STATUS_NOMEMORY;
    goto done;
  }
  pMapper->opaque.pDevObj = pDevObj;
  pMapper->opaque.maxEntry = maxEntry;
  for (i = 0; i < maxEntry; i++) {
    pMapper->opaque.bounce[i].pBouncePage = NULL;
  }
  pMapper->opaque.state.mapIndex = 0;
  pMapper->opaque.state.unmapIndex = 0;

  for (i = 0; i < maxEntry; i++) {
    /* get_zeroed_page() guarantees returned buffer is not in high memory? */
    pMapper->opaque.bounce[i].pBouncePage = virt_to_page(get_zeroed_page(GFP_KERNEL));
    if (NULL == pMapper->opaque.bounce[i].pBouncePage) {
      status = OS_STATUS_NOMEMORY;
      goto done;
    }
  }

 done:
  if (!OS_SUCCESS(status)) {
    if (NULL != pMapper) {
      for (i = 0; i < maxEntry; i++) {
	if (NULL == pMapper->opaque.bounce[i].pBouncePage) {
	  break;
	}
	free_page((unsigned long)page_address(pMapper->opaque.bounce[i].pBouncePage));
      }
      kfree(pMapper);
    }
  } else {
    *ppMapper = pMapper;
  }
  return status;
}

void
os_dma_mapper_destroy(
  OS_DMA_MAPPER* pMapper)
{
  unsigned int i, maxEntry;
  
  maxEntry = pMapper->opaque.maxEntry;
  for (i = 0; i < maxEntry; i++) {
    os_assert(NULL != pMapper->opaque.bounce[i].pBouncePage);
    if (NULL == pMapper->opaque.bounce[i].pBouncePage) {
      continue;
    }
    free_page((unsigned long)page_address(pMapper->opaque.bounce[i].pBouncePage));
  }
  kfree(pMapper);
}

OS_STATUS
os_usbd_create(
  void*         user,
  unsigned long length,
  OS_USBD**     usbd)
{
  OS_STATUS status = OS_STATUS_SUCCESS;
  OS_USBD* u = NULL;
  OS_USBD_REGION* region;
  struct page* pg;
  long nmapped = 0;
  unsigned long i, j, npage, nregion;
  unsigned long userl;
  unsigned long page_addr, page_limit_addr;
  unsigned long pg_remain;

  u = (OS_USBD*) os_kernel_malloc(sizeof(OS_USBD));
  if (u == NULL) {
    return OS_STATUS_NOMEMORY;
  }

  userl = (unsigned long)user;
  page_addr = userl & os_page_mask_high;
  page_limit_addr = (userl + length - 1) & os_page_mask_high;
  npage = ((page_limit_addr - page_addr) >> os_page_shift) + 1;
  nregion = (npage + OS_NPAGE_PER_REGION - 1) / OS_NPAGE_PER_REGION;

  u->user = user;
  u->length = length;
  u->first_page_offset = userl & os_page_mask_low;
  u->npage = npage;
  u->nregion = nregion;
  u->regions = (OS_USBD_REGION*)os_kernel_malloc(sizeof(OS_USBD_REGION) * nregion);
  if (u->regions == NULL) {
    status = OS_STATUS_NOMEMORY;
    goto failed;
  } else {
    for (i = 0; i < nregion; i++) {
      u->regions[i].pages = NULL;
    }
  }
  pg_remain = npage;
  for (i = 0; i < nregion; i++) {
    unsigned long n = (pg_remain > OS_NPAGE_PER_REGION) ? OS_NPAGE_PER_REGION : pg_remain;

    region = &u->regions[i];
    region->pages = (struct page**)os_kernel_malloc(n * sizeof(struct page*));
    if (region->pages == NULL) {
      status = OS_STATUS_NOMEMORY;
      goto failed;
    }
    for (j = 0; j < n; j++) {
      region->pages[j] = NULL;
    }

    down_read(&current->mm->mmap_sem);
    nmapped = get_user_pages(current,
			     current->mm,
			     page_addr,
			     n,
			     1, /* request write access */
			     0,
			     region->pages,
			     NULL);
    up_read(&current->mm->mmap_sem);
    if (nmapped < n) {
      os_dbgprint(1, ("*** os_create_usbd: get_user_pages returned %ld, va=0%lx n=%lu\n",
		      (long)nmapped, page_addr, n));
      status = OS_STATUS_UNSUCCESSFUL;
      goto failed;
    }

    page_addr += (n << os_page_shift);
    pg_remain -= n;
  }

  *usbd = u;

  return OS_STATUS_SUCCESS;

 failed:
  if (u != NULL) {
    if (u->regions != NULL) {
      pg_remain = u->npage;
      nregion = u->nregion;
      for (i = 0; i < nregion; i++) {
	unsigned long n = (pg_remain > OS_NPAGE_PER_REGION) ? OS_NPAGE_PER_REGION : pg_remain;

	region = &u->regions[i];
	if (region->pages != NULL) {
	  for (j = 0; j < n; j++) {
	    pg = region->pages[j];
	    if (pg == NULL) {
	      continue;
	    }
	    /* No need to set pages dirty because we didn't do anything with them */
	    page_cache_release(pg);
	  }
	  os_kernel_free(region->pages);
	}
	pg_remain -= n;
      }
      os_kernel_free(u->regions);
    }
    os_kernel_free(u);
  }
  return status;
}

os_boolean
os_delay_thread(
  os_boolean interruptible,
  os_ticks* ticks)
{
  unsigned long start, elapsed;
  unsigned long n = *ticks;

  if (interruptible) {
    set_current_state(TASK_INTERRUPTIBLE);
  } else {
    set_current_state(TASK_UNINTERRUPTIBLE);
  }

  start = jiffies;
  schedule_timeout(n);

  /* Calculate remaining time (this should work even if 'jiffies' wraps) */
  elapsed = jiffies - start;
  if (elapsed >= n) {
    *ticks = 0;
  } else {
    *ticks = n - elapsed;
  }

  if (interruptible && signal_pending(current)) {
    /* Caught a signal */
    return TRUE;
  }

  return FALSE;
}

void
os_usbd_destroy(
  OS_USBD* u)
{
  unsigned int i, j, pg_remain, nregion, n;
  OS_USBD_REGION* region;
  struct page* pg;

  if (u->regions != NULL) {
    pg_remain = u->npage;
    nregion = u->nregion;
    for (i = 0; i < nregion; i++) {
      n = (pg_remain > OS_NPAGE_PER_REGION) ? OS_NPAGE_PER_REGION : pg_remain;
      region = &u->regions[i];
      if (region->pages != NULL) {
	for (j = 0; j < n; j++) {
	  pg = region->pages[j];
	  if (pg == NULL) {
	    continue;
	  }
	  if (!PageReserved(pg)) {
	    SetPageDirty(pg);
	  }
	  page_cache_release(pg);
	}
	os_kernel_free(region->pages);
      }
      pg_remain -= n;
    }
    os_kernel_free(u->regions);
  }
  os_kernel_free(u);
}

void
os_disallow_mapping(
  OS_DEVICE_OBJECT*   pDevObj,
  OS_MAPPABLE_REGION* region)
{
  os_dbgprint(3, ("os_disallow_mapping: region = 0x%p\n", region));

  *region->prev = region->next;

  os_kernel_free(region);
}

void
os_usbd_flush(
  OS_USBD*      usbd,
  unsigned long offset,
  unsigned long length,
  os_boolean    write)
{
#ifdef CONFIG_NONCOHERENT_IO
  /*
  ** This is necessary because MIPS D-cache uses partial virtual address
  ** to generate cache line index; hence possibility of cache line aliasing
  ** where two or more cache lines may reference the same physical memory
  ** locations. To get around this problem we just flush the entire D-cache.
  */
  flush_cache_all();

#ifdef CONFIG_BOARD_SCACHE
  {
    unsigned long chunk;
    unsigned long p_idx, p_offs;
    unsigned long r_idx;
    os_uint8* kAddr;
    struct page* p;
    OS_USBD_REGION* region;
    struct page** pages;

    /*
    ** Make secondary cache coherent.
    **
    ** On MIPS, secondary cache is physically indexed; hence cache line
    ** aliasing is not a problem and we can selectively write-back & invalidate
    ** just the cache lines that are part of the buffer.
    */
    offset += usbd->first_page_offset;
    p_idx = offset >> os_page_shift;
    r_idx = p_idx >> OS_NPAGE_PER_REGION_ORDER;
    p_idx &= (OS_NPAGE_PER_REGION - 1);
    p_offs = offset & os_page_mask_low;
    region = &usbd->regions[r_idx];
    pages = region->pages + p_idx;
    while (length) {   
      p = *pages++;
      chunk = os_page_size - p_offs;
      if (chunk > length) {
	chunk = length;
      }
      kAddr = ((os_uint8*) page_address(p)) + p_offs;
      bc_wback_inv(kAddr, chunk);
      length -= chunk;
      p_idx++;
      if (p_idx == OS_NPAGE_PER_REGION) {
	p_idx = 0;
	r_idx++;
	region = &usbd->regions[r_idx];
	pages = region->pages;
      }
      p_offs = 0;
    }
  }
#endif /* def CONFIG_BOARD_SCACHE */
#endif /* def CONFIG_NONCOHERENT_IO */
}

void*
os_invoke_atomic(
  OS_ATTACHED_IRQ* obj,
  OS_INVOKE_FUNC*  function,
  void*            context)
{
  void*             ret;

  disable_irq(obj->number);
  ret = function(context);
  enable_irq(obj->number);

  return ret;
}

void*
os_map_kernel(
  OS_DEVICE_OBJECT* pDevObj,
  OS_DEVICE_RESOURCES* pResources,
  unsigned int barIndex,
  size_t length,
  OS_BAR_HANDLE* pHandle)
{
  void* p;

  if (!length) {
    os_dbgprint(0, ("os_map_kernel: attempt to map zero-length region\n"));
  }

  p = ioremap_nocache(pResources->translated.pci.base[barIndex], length);
  *pHandle = NULL;

  return p;
}

void
os_unmap_kernel(
  OS_DEVICE_OBJECT* pDevObj,
  OS_DEVICE_RESOURCES* pResources,
  unsigned int barIndex,
  size_t length,
  OS_BAR_HANDLE* pHandle,
  void* pKernelAddress)
{
  if (pKernelAddress == NULL) {
    os_dbgprint(0, ("os_unmap_kernel: NULL pointer\n"));
  }

  iounmap(pKernelAddress);
}

os_boolean
os_set_cancel_routine(
  OS_IO_REQUEST*    req,
  OS_IO_CANCELFUNC* func)
{
  OS_SPINLOCK_FLAGS f;

  os_acquire_spinlock(&req->opaque.pDevObj->cancelLock, &f);
  if (req->opaque.cancel) {
    /*
    ** Cannot change the cancel routine, as if set, it may be already
    ** running or about to run
    */
    os_release_spinlock(&req->opaque.pDevObj->cancelLock, f);
    return FALSE;
  } else {
    /*
    ** Can successfully change the cancel routine as if set, it's not
    ** running yet.
    */
    req->opaque.cancelFunc = func;
    os_release_spinlock(&req->opaque.pDevObj->cancelLock, f);
    return TRUE;
  }
}

/*
** --------------------------------------------------------------
** Event functions
** --------------------------------------------------------------
*/

void
os_init_event(
  OS_EVENT* event)
{
  os_init_spinlock(&event->lock);
  init_waitqueue_head(&event->q);
  event->bRaised = FALSE;
}

void
os_signal_event(
  OS_EVENT* event)
{
  OS_SPINLOCK_FLAGS f;

  os_acquire_spinlock(&event->lock, &f);
  os_assert(!event->bRaised);
  event->bRaised = TRUE;
  wake_up(&event->q);
  os_release_spinlock(&event->lock, f);
}

void
os_reset_event(
  OS_EVENT* event)
{
  OS_SPINLOCK_FLAGS f;

  os_acquire_spinlock(&event->lock, &f);
  event->bRaised = FALSE;
  os_release_spinlock(&event->lock, f);
}

os_boolean
os_wait_event(
  OS_EVENT*  event,
  os_boolean interruptible)
{
  OS_SPINLOCK_FLAGS f;
  wait_queue_t wait;

  os_acquire_spinlock(&event->lock, &f);
  if (event->bRaised) {
    os_release_spinlock(&event->lock, f);
    return FALSE;
  }
  if (interruptible) {
    set_current_state(TASK_INTERRUPTIBLE);
  } else {
    set_current_state(TASK_UNINTERRUPTIBLE);
  }
  init_waitqueue_entry(&wait, current);
  add_wait_queue(&event->q, &wait);
  os_release_spinlock(&event->lock, f);

  schedule();

  remove_wait_queue(&event->q, &wait);

  if (interruptible) {
    if (signal_pending(current)) {
      /* Caught a signal */
      return TRUE;
    }
  }

  /* No signal caught */
  return FALSE;
}

os_boolean
os_wait_event_timeout(
  OS_EVENT* event,
  os_ticks timeout_ticks,
  os_boolean bInterruptible,
  os_boolean* bSignalled)
{
  OS_SPINLOCK_FLAGS f;
  wait_queue_t wait;
  os_ticks expire;

  expire = os_get_time() + timeout_ticks;
  os_acquire_spinlock(&event->lock, &f);
  if (event->bRaised) {
    os_release_spinlock(&event->lock, f);
    /* Return without modifying timeout value */
    return FALSE;
  }
  if (bInterruptible) {
    set_current_state(TASK_INTERRUPTIBLE);
  } else {
    set_current_state(TASK_UNINTERRUPTIBLE);
  }
  init_waitqueue_entry(&wait, current);
  add_wait_queue(&event->q, &wait);
  os_release_spinlock(&event->lock, f);

  schedule_timeout(timeout_ticks);
  remove_wait_queue(&event->q, &wait);

  if (bInterruptible && signal_pending(current)) {
    *bSignalled = TRUE;
    return TRUE;
  } else {
    *bSignalled = FALSE;
  }

  if (time_after_eq((unsigned long)jiffies, (unsigned long)expire)) {
    /* Timeout */
    return TRUE;
  }

  return FALSE;
}

/*
** --------------------------------------------------------------
** Interrupt functions
** --------------------------------------------------------------
*/

OS_STATUS
os_attach_interrupt(
  OS_DEVICE_OBJECT* pDevObj,
  OS_ISR*           isr,
#if OS_ATTACHED_IRQ_HAS_DPC
  OS_DPC_FUNC*      dsr,
#endif
  OS_DEVICE_RESOURCES* pResources,
  const char*       devName,
  OS_ATTACHED_IRQ** o)
{
  OS_ATTACHED_IRQ* obj;
  int              err;

  obj = os_kernel_malloc(sizeof(OS_ATTACHED_IRQ));
  if (obj == NULL) {
    return OS_STATUS_NOMEMORY;
  }
  obj->number = pResources->translated.pci.irqNumber;
  obj->isr = isr;
  obj->pDevObj = pDevObj;
  obj->deviceContext = pDevObj->deviceContext;
#if OS_ATTACHED_IRQ_HAS_DPC
  os_init_dpc(&obj->dpc, pDevObj, dsr, obj->deviceContext);
#endif
#if (LINUX_VERSION_CODE >= 0x020612)
  err = request_irq(obj->number, _os_isr, IRQF_SHARED, devName, (void*) obj);
#else
  err = request_irq(obj->number, _os_isr, SA_SHIRQ | SA_INTERRUPT, devName, (void*) obj);
#endif
  if (err == 0) {
    *o = obj;
    return OS_STATUS_SUCCESS;
  } else {
    os_kernel_free(obj);
    if (err == -EBUSY) {
      return OS_STATUS_CANNOTSHARE;
    } else {
      return OS_STATUS_UNSUCCESSFUL;
    }
  }
}

void
os_detach_interrupt(
  OS_ATTACHED_IRQ* obj)
{
  free_irq(obj->number, obj);
  os_kernel_free(obj);
}

/*
** --------------------------------------------------------------
** DPC functions
** --------------------------------------------------------------
*/

static void
os_init_dpc(
  OS_DPC*           pDpc,
  OS_DEVICE_OBJECT* pDevObj,
  OS_DPC_FUNC*      pHandler,
  void*             pContext)
{
  pDpc->handler = pHandler;
  pDpc->context = pContext;
  tasklet_init(&pDpc->tasklet, _os_dpc_handler, (unsigned long)pDpc);
}

OS_STATUS
os_create_dpc(
  OS_DEVICE_OBJECT* pDevObj,
  OS_DPC_FUNC*      handler,
  void*             context,
  OS_DPC**          dpc)
{
  OS_DPC* p;

  p = (OS_DPC*)os_kernel_malloc(sizeof(OS_DPC));
  if (p == NULL) {
    return OS_STATUS_NOMEMORY;
  }
  os_init_dpc(p, pDevObj, handler, context);
  *dpc = p;
  return OS_STATUS_SUCCESS;
}

void
os_destroy_dpc(
  OS_DPC* dpc)
{
  tasklet_kill(&dpc->tasklet);
  os_kernel_free(dpc);
}

OS_STATUS
os_schedule_dpc(
  OS_DPC* dpc,
  void* pArgument)
{
  dpc->argument = pArgument;
  tasklet_schedule(&dpc->tasklet);
  
  return OS_STATUS_SUCCESS;
}

/*
** --------------------------------------------------------------
** Lookaside cache functions
** --------------------------------------------------------------
*/

OS_STATUS
os_create_lcache(
  unsigned long item_size,
  int           flags,
  OS_LCACHE**   cache)
{
  char          x[32];
  unsigned long f;
  OS_LCACHE*    c;

  c = (OS_LCACHE*) os_kernel_malloc(sizeof(OS_LCACHE));
  c->name[0] = '\0';
  c->cache = NULL;

  if (driver_name != NULL) {
    strncpy(c->name, driver_name, 10);
  }
  sprintf(x, "_%ld", (unsigned long) lcache_index);
  strcat(c->name, x);
  lcache_index++;
  
  f = SLAB_HWCACHE_ALIGN;
  os_dbgprint(2, ("os_create_lcache: creating kmem_cache, name=%s item_size=%ld\n",
		  c->name, (unsigned long) item_size));
#if KMEM_CACHE_CREATE_5
  /* In Linux 2.6.23 the destructor parameter disappeared */
  c->cache = kmem_cache_create(c->name, item_size, 0, f, NULL);
#else
  c->cache = kmem_cache_create(c->name, item_size, 0, f, NULL, NULL);
#endif
  if (c->cache == NULL) {
    os_kernel_free(c);
    return OS_STATUS_NOMEMORY;
  }

  *cache = c;
  return OS_STATUS_SUCCESS;
}

void
os_destroy_lcache(
  OS_LCACHE* cache)
{
  if (cache != NULL) {
    if (cache->cache != NULL) {
#if KMEM_CACHE_DESTROY_INT
      if (kmem_cache_destroy(cache->cache)) {
	os_dbgprint(0, ("*** os_destroy_lcache: memory leak detected\n"));
      }
#else
      kmem_cache_destroy(cache->cache);
#endif
      cache->cache = NULL;
    }
    os_kernel_free(cache);
  } else {
    os_dbgprint(0, ("*** os_destroy_lcache: NULL pointer\n"));
    return;
  }
}

/*
** --------------------------------------------------------------
** Memory allocation/freeing
** --------------------------------------------------------------
*/

void* os_kernel_malloc_dma(
  OS_DEVICE_OBJECT* pDevObj,
  unsigned long size,
  os_uint64* logicalAddr)
{
  os_uint64 log;
  dma_addr_t da;
  void* p;

  p = pci_alloc_consistent(pDevObj->dev.pci, size, &da);
  if (p == NULL) {
    return NULL;
  }
  log = (os_uint64)da;
  *logicalAddr = log;
  os_dbgprint(9, ("os_kernel_malloc_dma: size=%ld kern=0x%p log=0x%8.8lx%8.8lx\n",
                  size, p, (unsigned long)(log >> 32), (unsigned long)(os_uint32)log));
  return p;
}

void os_kernel_free_dma(
  OS_DEVICE_OBJECT* pDevObj,
  unsigned long size,
  void* kernelAddr,
  os_uint64 logicalAddr)
{
  dma_addr_t da;

  da = (dma_addr_t)logicalAddr;
  pci_free_consistent(pDevObj->dev.pci, size, kernelAddr, da);
}
